@@ -37,6 +37,7 @@ import sys
 import gzip
 import shutil
 import json
+from pathlib import Path
 from sklearn.neighbors import BallTree
 from geograpy.wikidata import Wikidata
 from lodstorage.sql import SQLDB
@@ -103,6 +104,7 @@ class LocationList(JSONAbleList):
             if 'wikidataid' in location.__dict__:
                 if location.wikidataid == wikidataID:
                     return location
+        return None
 
 
     @staticmethod
@@ -111,6 +113,45 @@ class LocationList(JSONAbleList):
             content = urlResponse.read().decode()
             return content
 
+    @staticmethod
+    def getFileContent(path:str):
+        with open(path, "r") as file:
+            content=file.read()
+            return content
+
+    @staticmethod
+    def getBackupDirectory():
+        home = str(Path.home())
+        path = f"{home}/.geograpy3"
+        return path
+
+    @staticmethod
+    def downloadBackupFile(url:str, fileName:str):
+        '''
+        Downloads from the given url the zip-file and extracts the file corresponding to the given fileName.
+
+        Args:
+            url: url linking to a downloadable gzip file
+            fileName: Name of the file that should be extracted from gzip file
+
+        Returns:
+            Name of the extracted file with path to the backup directory
+        '''
+        backupDirectory=LocationList.getBackupDirectory()
+        if not os.path.isdir(backupDirectory):
+            os.makedirs(backupDirectory)
+        extractTo= f"{backupDirectory}/{fileName}"
+        zipped = f"{extractTo}.gz"
+        print(f"Downloading {zipped} from {url} ... this might take a few seconds")
+        urllib.request.urlretrieve(url, zipped)
+        print(f"unzipping {extractTo} from {zipped}")
+        with gzip.open(zipped, 'rb') as gzipped:
+            with open(extractTo, 'wb') as unzipped:
+                shutil.copyfileobj(gzipped, unzipped)
+        if not os.path.isfile(extractTo):
+            raise (f"could extract {fileName} from {zipped}")
+        return extractTo
+
 
 class CountryList(LocationList):
     '''
@@ -189,8 +230,10 @@ class CountryList(LocationList):
             CountryList based on the json backup
         '''
         countryList = CountryList()
-        url="https://raw.githubusercontent.com/wiki/somnathrakshit/geograpy3/data/countries_geograpy3.json"
-        jsonStr = LocationList.getURLContent(url)
+        fileName="countries_geograpy3.json"
+        url="https://raw.githubusercontent.com/wiki/somnathrakshit/geograpy3/data/countries_geograpy3.json.gz"
+        backupFile=LocationList.downloadBackupFile(url, fileName)
+        jsonStr = LocationList.getFileContent(backupFile)
         countryList.restoreFromJsonStr(jsonStr)
         return countryList
 
@@ -285,8 +328,10 @@ class RegionList(LocationList):
             RegionList based on the json backup
         '''
         regionList = RegionList()
-        url = "https://raw.githubusercontent.com/wiki/somnathrakshit/geograpy3/data/regions_geograpy3.json"
-        jsonStr=LocationList.getURLContent(url)
+        fileName="regions_geograpy3.json"
+        url = "https://raw.githubusercontent.com/wiki/somnathrakshit/geograpy3/data/regions_geograpy3.json.gz"
+        backupFile = LocationList.downloadBackupFile(url, fileName)
+        jsonStr = LocationList.getFileContent(backupFile)
         regionList.restoreFromJsonStr(jsonStr)
         return regionList
 
@@ -322,67 +367,86 @@ class CityList(LocationList):
         for cityRecord in cityLOD:
             if 'city' in cityRecord:
                 wikidataid = Wikidata.getWikidataId(cityRecord['city'])
-                if wikidataid in cityIDs:
-                    # city already in the cityList -> merge
-                    city = cityList.getLocationByID(wikidataid)
-                    # current assumption is that only population and label are duplicates
-                    if 'labels' in cityRecord:
-                        if 'labels' in city.__dict__:
-                            if isinstance(city.labels, list):
-                                if cityRecord['labels'] in city.labels:
-                                    city.labels.append(cityRecord['labels'])
-                            else:
-                                labels = []
-                                labels.append(city.labels)
-                                labels.append(cityRecord['labels'])
-                                city.labels = labels
-                        else:
-                            city.__dict__['labels'] = [cityRecord['labels']]
-                    if 'cityPop' in cityRecord:
-                        population = None
-                        if 'population' in city.__dict__:
-                            population = max(cityRecord['cityPop'], city.population)
-                        else:
-                            population = cityRecord['cityPop']
-                        city.population = population
-                else:
-                    # add new city to list
-                    city=City()
-                    city.wikidataid = wikidataid
-                    city.name = cityRecord['cityLabel']
-                    regionIso = cityRecord['regionIsoCode']
-                    city.partOf = regionIso
-                    if 'countryCoord' in cityRecord:
-                        lon, lat = Wikidata.getCoordinateComponents(cityRecord['cityCoord'])
-                        city.lat = lat
-                        city.lon = lon
-                    if 'country' in cityRecord:
-                        country = cityRecord['country']
-                        country_wikidataid = Wikidata.getWikidataId(country)
-                        city.country_wikidataid = country_wikidataid
-                    if 'region' in cityRecord:
-                        region = cityRecord['region']
-                        region_wikidataid = Wikidata.getWikidataId(region)
-                        city.region_wikidataid = region_wikidataid
-                    if 'cityPop' in cityRecord:
-                        city.population = cityRecord['cityPop']
-                    cityIDs.append(wikidataid)
-                    if cityList.listName not in cityList.__dict__ or cityList.__dict__[cityList.listName] is None:
-                        cityList.__dict__[cityList.listName]=[]
-                    cityList.__dict__[cityList.listName].append(city)
+                cityList.updateCity(wikidataid, cityRecord)
         return cityList
 
+    def updateCity(self, wikidataid:str, cityRecord:dict):
+        '''
+        Updates the city corresponding to the given city with the given data.
+        If the city does not exist a new city object is created and added to this CityList
+        Args:
+            wikidataid(str): wikidata id of the city that should be updated/added
+            cityRecord(dict): data of the given city that should be updated/added
+
+        Returns:
+            Nothing
+        '''
+        city = self.getLocationByID(wikidataid)
+        if city is not None:
+            # city already in the cityList -> merge
+            # current assumption is that only population and label are duplicates
+            if 'labels' in cityRecord:
+                if 'labels' in city.__dict__:
+                    if isinstance(city.labels, list):
+                        if cityRecord['labels'] in city.labels:
+                            city.labels.append(cityRecord['labels'])
+                    else:
+                        labels = []
+                        labels.append(city.labels)
+                        labels.append(cityRecord['labels'])
+                        city.labels = labels
+                else:
+                    city.__dict__['labels'] = [cityRecord['labels']]
+            if 'cityPop' in cityRecord:
+                population = None
+                if 'population' in city.__dict__  and city.population is not None:
+                    if cityRecord['cityPop'] is None:
+                        population=city.population
+                    else:
+                        population = max(float(cityRecord['cityPop']), float(city.population))
+                else:
+                    population = cityRecord['cityPop']
+                city.population = population
+        else:
+            # add new city to list
+            city = City()
+            city.wikidataid = wikidataid
+            city.name = cityRecord['cityLabel']
+            if 'countryCoord' in cityRecord:
+                lon, lat = Wikidata.getCoordinateComponents(cityRecord['cityCoord'])
+                city.lat = lat
+                city.lon = lon
+            if 'country' in cityRecord:
+                country = cityRecord['country']
+                country_wikidataid = Wikidata.getWikidataId(country)
+                city.country_wikidataid = country_wikidataid
+            if 'region' in cityRecord:
+                region = cityRecord['region']
+                region_wikidataid = Wikidata.getWikidataId(region)
+                city.region_wikidataid = region_wikidataid
+            if 'cityPop' in cityRecord:
+                city.population = cityRecord['cityPop']
+            if self.listName not in self.__dict__ or self.__dict__[self.listName] is None:
+                self.__dict__[self.listName] = []
+            self.__dict__[self.listName].append(city)
+
     @classmethod
-    def fromJSONBackup(cls):
+    def fromJSONBackup(cls, jsonStr:str=None):
         '''
         get city list from json backup (json backup is based on wikidata query results)
 
+        Args:
+            jsonStr(str): JSON string the CityList should be loaded from. If None json backup is loaded. Default is None
+
         Returns:
             CityList based on the json backup
         '''
         cityList = CityList()
-        url = "https://raw.githubusercontent.com/wiki/somnathrakshit/geograpy3/data/cities_geograpy3.json"
-        jsonStr = LocationList.getURLContent(url)
+        if jsonStr is None:
+            fileName="cities_geograpy3.json"
+            url = "https://raw.githubusercontent.com/wiki/somnathrakshit/geograpy3/data/cities_geograpy3.json.gz"
+            backupFile = LocationList.downloadBackupFile(url, fileName)
+            jsonStr = LocationList.getFileContent(backupFile)
         cityList.restoreFromJsonStr(jsonStr)
         return cityList
     
@@ -545,7 +609,9 @@ class City(Location):
             "level": 5,
             "locationKind": "City",
             "comment": None,
-            "population": "3976322"
+            "population": "3976322",
+            "region_wikidataid": "Q99",
+            "country_wikidataid": "Q30"
         }]
         return samplesLOD
     
@@ -622,7 +688,8 @@ class Region(Location):
             "locationKind": "Region",
             "comment": None,
             "labels": ["CA", "California"],
-            "iso": "US-CA"
+            "iso": "US-CA",
+            "country_wikidataid": "Q30"
         }]
         return samplesLOD
     
@@ -1116,7 +1183,7 @@ class Locator(object):
             self.populate_Version(self.sqlDB)
     
         elif not hasData:
-            url="http://wiki.bitplan.com/images/confident/locs.db.gz"
+            url="https://github.com/somnathrakshit/geograpy3/releases/download/0.1.27/locs.db.gz"
             zipped=self.db_file+".gz"
             print("Downloading %s from %s ... this might take a few seconds" % (zipped,url))
             urllib.request.urlretrieve(url,zipped)
@@ -1214,7 +1281,7 @@ FROM City_wikidata
         dbFile=self.db_path+"/City_wikidata.db"
         if not os.path.exists(dbFile):
             print("Downloading %s ... this might take a few seconds" % dbFile)
-            dbUrl="http://wiki.bitplan.com/images/confident/City_wikidata.db"
+            dbUrl="https://github.com/somnathrakshit/geograpy3/releases/download/0.1.27/City_wikidata.db"
             urllib.request.urlretrieve(dbUrl,dbFile)
         wikiCitiesDB=SQLDB(dbFile)
         wikiCitiesDB.copyTo(sqlDB)
@@ -1240,7 +1307,7 @@ FROM City_wikidata
             else:
                 # just download a copy 
                 print("Downloading %s ... this might take a few seconds" % dbFile)
-                dbUrl="http://wiki.bitplan.com/images/confident/city_wikidata_population.db"
+                dbUrl="https://github.com/somnathrakshit/geograpy3/releases/download/0.1.27/city_wikidata_population.db"
                 urllib.request.urlretrieve(dbUrl,dbFile)
         # (re) open the database
         wikiCitiesDB=SQLDB(dbFile) 
